#!/usr/bin/env Rscript

# parse assembly metadata for NCBI genomes

### setting up environment ####

# Check if packages are installed
package.list <- c(
  "crayon",
  "optparse",
  "config"
)

# Function to check if packages are installed
is.installed <- function(pkg){
  is.element(pkg, installed.packages()[,1])
}

# If not all packages are available
if(any(!is.installed(package.list))) {
  cat("Not all required packages are available. They will now be installed.\n")
  
  # give the user the chance to abort manually
  Sys.sleep(20)
  
  # then install packages
  for(i in which(!is.installed(package.list))) {
    suppressMessages(install.packages(package.list[i], repos = "http://cran.us.r-project.org"))
  }
}

# Break the script if the package installation was unsuccessful
if(any(!is.installed(package.list))) {
  cat(
    paste0(
      "Unable to install all required packages.\nPlease install ",
      paste0(package.list[!is.installed(package.list)], collapse = ", "),
      " manually."
    )
  )
  break
}

# Load packages
cat("Loading libraries...")
silent <- suppressMessages(lapply(package.list, function(X) {require(X, character.only = TRUE)}))
rm(silent)
cat(" done\n")

# Some functions for message output
msg <- function(X){
  cat(crayon::white(paste0("[",format(Sys.time(), "%T"), "]")), X)
}
msg_sub <- function(X){
  cat(crayon::white(paste0("  [",format(Sys.time(), "%T"), "]")), X)
}


### Reading command line options ####

# define command line options
option_list <- list(
  make_option(
    c("-f", "--fr_R1"),
    type = "character",
    default = NULL,
    help = "md5sums of fr_R1 read file",
    metavar = "character"
  ),
  make_option(
    c("-F", "--fr_R2"),
    type = "character",
    default = NULL,
    help = "md5sums of fr_R2 read file",
    metavar = "character"
  ),
  make_option(
    c("-r", "--rf_R1"),
    type = "character",
    default = NULL,
    help = "md5sums of rf_R1 read file",
    metavar = "character"
  ),
  make_option(
    c("-R", "--rf_R2"),
    type = "character",
    default = NULL,
    help = "md5sums of rf_R2 read file",
    metavar = "character"
  ),
  make_option(
    c("-s", "--sample_list"),
    type = "character",
    default = NULL,
    help = "file name for sample list",
    metavar = "character"
  ),
  make_option(
    c("-o", "--output"),
    type = "character",
    default = NULL,
    help = "name of output file",
    metavar = "character"
  )
)

opt_parser <- OptionParser(option_list = option_list)
opt <- parse_args(opt_parser)

if (is.null(opt$fr_R1) | is.null(opt$fr_R2) | is.null(opt$rf_R1) |
    is.null(opt$rf_R2) | is.null(opt$sample_list) | is.null(opt$output)) {
  print_help(opt_parser)
  stop("All parameters are mandatory.\n", call. = FALSE)
}


### sorting md5 data ####

sample.names <- read.table(opt$sample_list, h = F, sep = "\t", stringsAsFactors = F)[, 1]
fr_R1 <- read.table(opt$fr_R1, h = F, stringsAsFactors = F)
fr_R2 <- read.table(opt$fr_R2, h = F, stringsAsFactors = F)
rf_R1 <- read.table(opt$rf_R1, h = F, stringsAsFactors = F)
rf_R2 <- read.table(opt$rf_R2, h = F, stringsAsFactors = F)

rownames(fr_R1) <- gsub("_clip_fr_R1.fastq.gz", "", basename(fr_R1$V2), fixed = T)
rownames(fr_R2) <- gsub("_clip_fr_R2.fastq.gz", "", basename(fr_R2$V2), fixed = T)
rownames(rf_R1) <- gsub("_clip_rf_R1.fastq.gz", "", basename(rf_R1$V2), fixed = T)
rownames(rf_R2) <- gsub("_clip_rf_R2.fastq.gz", "", basename(rf_R2$V2), fixed = T)

fr_R1 <- fr_R1[sample.names, ]
fr_R2 <- fr_R2[sample.names, ]
rf_R1 <- rf_R1[sample.names, ]
rf_R2 <- rf_R2[sample.names, ]

checksums <- data.frame(
  seq_id = sample.names,
  forward_read_file_name = paste0(
    basename(fr_R1$V2),
    ";",
    basename(rf_R1$V2)
  ),
  forward_read_file_checksum = paste0(
    basename(fr_R1$V1),
    ";",
    basename(rf_R1$V1)
  ),
  reverse_read_file_name = paste0(
    basename(fr_R2$V2),
    ";",
    basename(rf_R2$V2)
  ),
  reverse_read_file_checksum = paste0(
    basename(fr_R2$V1),
    ";",
    basename(rf_R2$V1)
  )
)

write.table(
  checksums,
  opt$output,
  sep = "\t",
  quote = F,
  row.names = F
)

