#!/usr/bin/env Rscript

# parse assembly metadata for NCBI genomes

### setting up environment ####

# Check if packages are installed
package.list <- c(
  "crayon",
  "optparse",
  "config",
  "dada2",
  "data.table",
  "tidyverse"
)

# Function to check if packages are installed
is.installed <- function(pkg){
  is.element(pkg, installed.packages()[,1])
}

# If not all packages are available
if(any(!is.installed(package.list))) {
  cat("Not all required packages are available. They will now be installed.\n")
  
  # give the user the chance to abort manually
  Sys.sleep(20)
  
  # then install packages
  for(i in which(!is.installed(package.list))) {
    suppressMessages(install.packages(package.list[i], repos = "http://cran.us.r-project.org"))
  }
}

# Break the script if the package installation was unsuccessful
if(any(!is.installed(package.list))) {
  cat(
    paste0(
      "Unable to install all required packages.\nPlease install ",
      paste0(package.list[!is.installed(package.list)], collapse = ", "),
      " manually."
    )
  )
  break
}

# Load packages
cat("Loading libraries...")
silent <- suppressMessages(lapply(package.list, function(X) {require(X, character.only = TRUE)}))
rm(silent)
cat(" done\n")

# Some functions for message output
msg <- function(X){
  cat(crayon::white(paste0("[",format(Sys.time(), "%T"), "]")), X)
}
msg_sub <- function(X){
  cat(crayon::white(paste0("  [",format(Sys.time(), "%T"), "]")), X)
}


### Reading command line options ####

# define command line options
option_list <- list(
  make_option(
    c("-n", "--name"),
    type = "character", 
    default = NULL, 
    help = "project name (will serve as file base name for output files",
    metavar = "character"
  ),
  make_option(
    c("-a", "--asv_table"),
    type = "character",
    default = NULL,
    help = "file name of of ASV table (sequences are rownames)",
    metavar = "character"
  ),
  make_option(
    c("-b", "--bootstrap"),
    type = "integer",
    default = NULL, 
    help = "minimum bootstrap value to trust dada2 classification", 
    metavar = "number"
  ),
  make_option(
    c("-t", "--train_set"),
    type = "character",
    default = NULL,
    help = "location of reference training data set for classification",
    metavar = "character"
  ),
  make_option(
    c("-l", "--lineage"),
    type = "character",
    default = "Bacteria,Archaea",
    help = "which domains of life were targeted (comma separated)",
    metavar = "character"
  ),
  make_option(
    c("-d", "--dir"),
    type = "character",
    default = NULL,
    help = "path to analysis directory (highest level directory for sequencing project",
    metavar = "character"
  ),
  make_option(
    c("-v", "--analysis"),
    type = "character",
    default = NULL,
    help = "name of the analysis (version) that is being run",
    metavar = "character"
  ),
  make_option(
    c("-c", "--cpus"),
    type = "integer",
    default = 1, 
    help = "number of threads [default: 1]", 
    metavar = "number"
  )
)

opt_parser <- OptionParser(option_list = option_list)
opt <- parse_args(opt_parser)

if (is.null(opt$name) | is.null(opt$asv_table) | is.null(opt$bootstrap) |
    is.null(opt$train_set) | is.null(opt$dir) | is.null(opt$analysis)) {
  print_help(opt_parser)
  stop("All parameters are mandatory.\n", call. = FALSE)
}


### Defining functions ####

# Further curation of taxonomic paths: removing NAs and uncultured
# append unclassified to last classified level instead of NA
curate_taxpath <- function(tax) {
  out <- tax
  out[out == "uncultured"] <- NA
  k <- ncol(out) - 1
  for (i in 2:k) {
    if (sum(is.na(out[, i])) > 1) {
      test <- out[is.na(out[, i]), ]
      for (j in 1:nrow(test)) {
        if (sum(is.na(test[j, i:(k + 1)])) == length(test[j, i:(k + 1)])) {
          test[j, i] <- paste(test[j, (i - 1)], "_unclassified", sep = "")
          test[j, (i + 1):(k + 1)] <- test[j, i]
        }
      }
      out[is.na(out[, i]), ] <- test
    }
    if (sum(is.na(out[, i])) == 1) {
      test <- out[is.na(out[, i]), ]
      if (sum(is.na(test[i:(k + 1)])) == length(test[i:(k + 1)])) {
        test[i] <- paste(test[(i - 1)], "_unclassified", sep = "")
        test[(i + 1):(k + 1)] <- test[i]
      }
      out[is.na(out[, i]),] <- test
    }
  }
  out[is.na(out[, (k + 1)]), (k + 1)] <- paste(out[is.na(out[, (k + 1)]), k], "_unclassified", sep = "")
  return(out)
}


### Read AVS table ####

# read ASV table
asv <- read.table(
  paste0(opt$dir, "/Intermediate_results/", opt$analysis, "/asvtab_", opt$name, ".txt"), 
  h = T,
  sep = "\t",
  stringsAsFactors = F, 
  row.names = 1, 
  check.names = FALSE
)


### taxonomic classification using RDP within dada2 ####

msg("Classifying with assignTaxonomy()...\n")

tax <- assignTaxonomy(
  t(asv), 
  opt$train_set,
  multithread = opt$cpus,
  minBoot = 0,
  outputBootstraps = T,
  taxLevels = c("Kingdom", "Phylum", "Class", "Order", "Family", "Genus")
)

# Remove unwanted lineages
# always remove chloroplast and mitochondrial sequences
# while these can be meaningful for the analysis, they are not part of this standard workflow
keep_lineage <- strsplit(opt$lineage, ",")[[1]]
tax.good <- lapply(tax, function(x) {
  x[tax$tax[, 1] %in% keep_lineage & 
      !grepl("[Cc]hloroplast", tax$tax[, 4]) &
      !grepl("[Mm]itochondria", tax$tax[, 5])
    , ] 
})
msg(paste0("Removing unwanted lineages resulted in ", round(nrow(tax.good$tax)/nrow(asv) * 100), "% of ASVs being retained.\n"))

# Overview of raw bootstrap values
msg("Summary of raw bootstrap values at each taxonomic level:\n")
apply(tax.good$boot, 2, summary)

# Applying bootstrap filter at pre-determined cut-off
tax.filt <- tax.good$tax
tax.filt[tax.good$boot < opt$bootstrap] <- NA
# remove all without kingdom assignment
tax.filt <- tax.filt[!is.na(tax.filt[, 1]), ]
msg("Number of unclasified ASVs at each taxonomic level:\n")
apply(tax.filt, 2, function(x) sum(is.na(x)))
msg(paste0("Applying a bootstrap cut-off of ", opt$bootstrap, " resulted in ", round(nrow(tax.filt)/nrow(tax.good$tax) * 100), "% of ASVs being retained.\n"))

# Which proportion of sequences is affected?
asv.filt <- asv[rownames(tax.filt), ]
msg("Summary of total sequence proportions [%] unclassified at each taxonomic level:\n")
tmp <- data.frame(apply(tax.filt, 2, function(x) sum(rowSums(asv.filt[is.na(x), ]))/sum(colSums(asv.filt))))
colnames(tmp) <- "percentage"
tmp
rm(tmp)

# Further curation of taxonomic paths
# append unclassified to last classified level instead of NA
tax.filt.clean <- curate_taxpath(tax.filt)

# change column names to lowercase for compatibility with blast output
colnames(tax.filt.clean) <- c("domain", "phylum", "class", "order", "family", "genus")

# write output file for dada2 classification
write.table(
  tax.filt.clean,
  paste0(opt$dir, "/Intermediate_results/", opt$analysis, "/tax_dada2_", opt$name, ".txt"),
  quote = F,
  sep = "\t"
)
save.image(paste0(opt$dir, "/Intermediate_results/", opt$analysis, "/workspace_classification_", opt$name, ".Rdata"))


