#!/usr/bin/env Rscript

# parse assembly metadata for NCBI genomes

### setting up environment ####

# Check if packages are installed
package.list <- c(
  "crayon",
  "optparse",
  "config",
  "dada2",
  "ShortRead",
  "ggplot2",
  "gridExtra",
  "Biostrings",
  "scales",
  "furrr"
)

# Function to check if packages are installed
is.installed <- function(pkg){
  is.element(pkg, installed.packages()[,1])
}

# If not all packages are available
if(any(!is.installed(package.list))) {
  cat("Not all required packages are available. They will now be installed.\n")
  
  # give the user the chance to abort manually
  Sys.sleep(20)
  
  # then install packages
  for(i in which(!is.installed(package.list))) {
    suppressMessages(install.packages(package.list[i], repos = "http://cran.us.r-project.org"))
  }
}

# Break the script if the package installation was unsuccessful
if(any(!is.installed(package.list))) {
  cat(
    paste0(
      "Unable to install all required packages.\nPlease install ",
      paste0(package.list[!is.installed(package.list)], collapse = ", "),
      " manually."
    )
  )
  break
}

# Load packages
cat("Loading libraries...")
silent <- suppressMessages(lapply(package.list, function(X) {require(X, character.only = TRUE)}))
rm(silent)
cat(" done\n")

# Some functions for message output
msg <- function(X){
  cat(crayon::white(paste0("[",format(Sys.time(), "%T"), "]")), X)
}
msg_sub <- function(X){
  cat(crayon::white(paste0("  [",format(Sys.time(), "%T"), "]")), X)
}


### Reading command line options ####

# define command line options
option_list <- list(
  make_option(
    c("-n", "--name"),
    type = "character",
    default = NULL,
    help = "project name (will serve as file base name for output files",
    metavar = "character"
  ),
  make_option(
    c("-s", "--samples"),
    type = "character",
    default = NULL,
    help = "file name of sample list (has to be in first column, no headers)",
    metavar = "character"
  ),
  make_option(
    c("-l", "--truncLen_R1"),
    type = "integer",
    default = NULL,
    help = "truncation length for R1 read",
    metavar = "number"
  ),
  make_option(
    c("-L", "--truncLen_R2"),
    type = "integer",
    default = NULL,
    help = "truncation length for R2 read",
    metavar = "number"
  ),
  make_option(
    c("-f", "--frag_min"),
    type = "integer",
    default = NULL,
    help = "minimum size of expected fragment",
    metavar = "number"
  ),
  make_option(
    c("-F", "--frag_max"),
    type = "integer",
    default = NULL,
    help = "maximum size of expected fragment",
    metavar = "number"
  ),
  make_option(
    c("-e", "--error_R1"),
    type = "integer",
    default = NULL,
    help = "maximum expected error rate R1 read",
    metavar = "number"
  ),
  make_option(
    c("-E", "--error_R2"),
    type = "integer",
    default = NULL,
    help = "maximum expected error rate R2 read",
    metavar = "number"
  ),
  make_option(
    c("-d", "--dir"),
    type = "character",
    default = NULL,
    help = "path to analysis directory (highest level directory for sequencing project",
    metavar = "character"
  ),
  make_option(
    c("-a", "--analysis"),
    type = "character",
    default = NULL,
    help = "name of the analysis (version) that is being run",
    metavar = "character"
  ),
  make_option(
    c("-c", "--cpus"),
    type = "integer",
    default = 1,
    help = "number of threads [default: 1]",
    metavar = "number"
  )
)

opt_parser <- OptionParser(option_list = option_list)
opt <- parse_args(opt_parser)

if (is.null(opt$name) | is.null(opt$samples) | is.null(opt$truncLen_R1) |
    is.null(opt$truncLen_R2) | is.null(opt$frag_min) | is.null(opt$frag_max) |
    is.null(opt$error_R1) | is.null(opt$error_R2) | is.null(opt$dir) | is.null(opt$analysis)) {
  print_help(opt_parser)
  stop("All parameters are mandatory.\n", call. = FALSE)
}


### Defining functions ####

# Wrapper function to create multi-page pdf with base quality plots
# Argument 'file_base' will be extended by '_R1.pdf' and '_R2.pdf'
quality_check <- function(
  R1, 
  R2, 
  file_base = "QualityProfile",
  cpus = 1
) {
  
  # start parallel session
  plan(multicore, workers = cpus)
  
  # R1 profiles
  QualityProfileFs <- future_map(
    1:length(R1),
    function(x) {
      tmp <- list()
      tmp[[1]] <- plotQualityProfile(R1[x])
      return(tmp)
    }
  )
  pdf(paste0(file_base, "_R1.pdf"))
  for(i in 1:length(R1)) {
    do.call("grid.arrange", QualityProfileFs[[i]])  
  }
  dev.off()
  
  # R2 profiles
  QualityProfileRs <- future_map(
    1:length(R2),
    function(x) {
      tmp <- list()
      tmp[[1]] <- plotQualityProfile(R2[x])
      return(tmp)
    }
  )
  pdf(paste0(file_base, "_R2.pdf"))
  for(i in 1:length(R2)) {
    do.call("grid.arrange", QualityProfileRs[[i]])  
  }
  dev.off()
}


### Defining input and output file paths ####

msg(paste0("This is dada2 version: ", packageVersion("dada2"), ".\n"))

# set sample names
sample.names <- read.table(opt$samples, h = F, sep = "\t", stringsAsFactors = F)[, 1]

# Specify path to input sequences
path <- paste0(opt$dir, "/Validated_data/", opt$analysis)
fnFs.fr <- paste0(path, "/", sample.names, "_clip_fr_R1.fastq.gz")
fnRs.fr <- paste0(path, "/", sample.names, "_clip_fr_R2.fastq.gz")
fnFs.rf <- paste0(path, "/", sample.names, "_clip_rf_R1.fastq.gz")
fnRs.rf <- paste0(path, "/", sample.names, "_clip_rf_R2.fastq.gz")
names(fnFs.fr) <- sample.names
names(fnRs.fr) <- sample.names
names(fnFs.rf) <- sample.names
names(fnRs.rf) <- sample.names

# Place filtered files in Filtered/ subdirectory
filtFs.fr <- file.path(opt$dir, "Intermediate_results", opt$analysis, "Filtered", paste0(sample.names, "_filt_fr_R1.fastq"))
filtRs.fr <- file.path(opt$dir, "Intermediate_results", opt$analysis, "Filtered", paste0(sample.names, "_filt_fr_R2.fastq"))
filtFs.rf <- file.path(opt$dir, "Intermediate_results", opt$analysis, "Filtered", paste0(sample.names, "_filt_rf_R1.fastq"))
filtRs.rf <- file.path(opt$dir, "Intermediate_results", opt$analysis, "Filtered", paste0(sample.names, "_filt_rf_R2.fastq"))
names(filtFs.fr) <- sample.names
names(filtRs.fr) <- sample.names
names(filtFs.rf) <- sample.names
names(filtRs.rf) <- sample.names


### Filter and trim reads with optimized parameters ####

# Run trimming with optimal parameters
filt.out.fr <- filterAndTrim(
  fwd = fnFs.fr, 
  filt = filtFs.fr, 
  rev = fnRs.fr, 
  filt.rev = filtRs.fr,
  truncLen = c(opt$truncLen_R1, opt$truncLen_R2),
  maxN = 0,
  minQ = 2,
  maxEE = c(opt$error_R1, opt$error_R2), 
  truncQ = 0, 
  rm.phix = TRUE,
  compress = F,
  multithread = opt$cpus
)
filt.out.rf <- filterAndTrim(
  fwd = fnFs.rf, 
  filt = filtFs.rf, 
  rev = fnRs.rf, 
  filt.rev = filtRs.rf,
  truncLen = c(opt$truncLen_R1, opt$truncLen_R2),
  maxN = 0,
  minQ = 2,
  maxEE = c(opt$error_R1, opt$error_R2), 
  truncQ = 0, 
  rm.phix = TRUE,
  compress = F,
  multithread = opt$cpus
)

# Repeat quality check after trimming
quality_check(
  c(filtFs.fr, filtFs.rf),
  c(filtRs.fr, filtRs.rf),
  file_base = paste0(opt$dir, "/Intermediate_results/", opt$analysis, "/QualityProfileFiltered_", opt$name),
  cpus = opt$cpus
)


### Denoising ####

# Learn error rates
errF.fr <- learnErrors(filtFs.fr, multithread = opt$cpus, randomize = TRUE, verbose = 1, MAX_CONSIST = 10)
errR.fr <- learnErrors(filtRs.fr, multithread = opt$cpus, randomize = TRUE, verbose = 1, MAX_CONSIST = 10)
errF.rf <- learnErrors(filtFs.rf, multithread = opt$cpus, randomize = TRUE, verbose = 1, MAX_CONSIST = 10)
errR.rf <- learnErrors(filtRs.rf, multithread = opt$cpus, randomize = TRUE, verbose = 1, MAX_CONSIST = 10)
save.image(paste0(opt$dir, "/Intermediate_results/", opt$analysis, "/workspace_denoising_", opt$name, ".Rdata"))

# check convergence of error estimation and plot error profiles
pdf(paste0(opt$dir, "/Intermediate_results/", opt$analysis, "/ErrorProfiles_", opt$name, ".pdf"))
barplot(log10(dada2:::checkConvergence(errF.fr) + 1), main = "Convergence_fwd")
barplot(log10(dada2:::checkConvergence(errR.fr) + 1), main = "Convergence_rev")
barplot(log10(dada2:::checkConvergence(errF.rf) + 1), main = "Convergence_fwd")
barplot(log10(dada2:::checkConvergence(errR.rf) + 1), main = "Convergence_rev")
plotErrors(errF.fr, nominalQ = TRUE)
plotErrors(errR.fr, nominalQ = TRUE)
plotErrors(errF.rf, nominalQ = TRUE)
plotErrors(errR.rf, nominalQ = TRUE)
dev.off()

# Dereplicate and denoise samples
dadaFs.fr <- dada(filtFs.fr, err = errF.fr, multithread = opt$cpus, pool = TRUE)
dadaRs.fr <- dada(filtRs.fr, err = errR.fr, multithread = opt$cpus, pool = TRUE)
dadaFs.rf <- dada(filtFs.rf, err = errF.rf, multithread = opt$cpus, pool = TRUE)
dadaRs.rf <- dada(filtRs.rf, err = errR.rf, multithread = opt$cpus, pool = TRUE)
save.image(paste0(opt$dir, "/Intermediate_results/", opt$analysis, "/workspace_denoising_", opt$name, ".Rdata"))


### Merging ####

# Merge reads
mergers.fr <- mergePairs(
  dadaFs.fr,
  filtFs.fr, 
  dadaRs.fr, 
  filtRs.fr, 
  minOverlap = 10,
  verbose = TRUE
)
mergers.rf <- mergePairs(
  dadaFs.rf,
  filtFs.rf, 
  dadaRs.rf, 
  filtRs.rf, 
  minOverlap = 10,
  verbose = TRUE
)

# Create sequence table
seqtab.fr <- makeSequenceTable(mergers.fr)
seqtab.rf <- makeSequenceTable(mergers.rf)
msg(paste0("There are ", ncol(seqtab.fr), " ASVs in fr orientation, and ", ncol(seqtab.rf), " ASVs in rf orientation after merging.\n"))

# This is the step at which separate denoising runs should be combined
# (e.g. if data comes from different sequencer runs or lanes, 
# or if fwd-rev and rev-fwd orientation were processed separately) 

# Generate reverse complement of rf
seqtab.rf.rc <- seqtab.rf
colnames(seqtab.rf.rc) <- rc(colnames(seqtab.rf))

# Merge sequence tables
seqtab <- mergeSequenceTables(
  seqtab.fr,
  seqtab.rf.rc,
  repeats = "sum"
)
msg(paste0("The combined fr and rf tables contain ", ncol(seqtab), " ASVs.\n"))


### Chimera removal and further cleaning steps ####

# Remove chimeras
# This may remove quite a bit of ASVs, but only a small fraction of your total sequences
seqtab.nochim <- removeBimeraDenovo(seqtab, method = "consensus", multithread = opt$cpus, verbose = TRUE)
msg(paste0(round(ncol(seqtab.nochim)/ncol(seqtab) * 100), "% of ASVs retained after chimera detection.\n"))
msg("Summary of retained sequences [%] per sample:\n")
round(summary(rowSums(seqtab.nochim)/rowSums(seqtab)) * 100)

# Sequence length distribution
write.table(
  data.frame(table(rep(nchar(colnames(seqtab.nochim)), colSums(seqtab.nochim)))),
  paste0(opt$dir, "/Intermediate_results/", opt$analysis, "/Length_distribution_", opt$name, ".txt"),
  sep = "\t",
  quote = F,
  row.names = F
)

# Remove potential junk sequences and singletons
# dada does not generate singletons, any singletons are introduced in the merging step
# Adjust range of sequence lengths based on expected length of marker gene fragment
seqtab.nochim2 <- seqtab.nochim[, nchar(colnames(seqtab.nochim)) %in% seq(opt$frag_min, opt$frag_max) & colSums(seqtab.nochim) > 1]
msg(paste0("There are ", ncol(seqtab.nochim2), " ASVs in the cleaned sequence count table.\n"))
msg(paste0("This corresponds to ", round(ncol(seqtab.nochim2)/ncol(seqtab) * 100), "% of ASVs compared to the merged data set.\n"))
msg("Summary of retained sequences [%] per sample:\n")
round(summary(rowSums(seqtab.nochim)/rowSums(seqtab)) * 100)


### Number of sequences at each analysis step ####

# Get number of sequences at each analysis step
nSeq <- read.table(paste0(opt$dir, "/Intermediate_results/", opt$analysis, "/nseqs_", opt$name, ".txt"), h = T, stringsAsFactors = F, row.names = 1)
nSeq <- nSeq[sample.names, ]

getN <- function(x) sum(getUniques(x))
track <- cbind(
  nSeq$Demux,
  filt.out.fr[, 1],
  filt.out.rf[, 1],
  filt.out.fr[, 2],
  filt.out.rf[, 2],
  sapply(dadaFs.fr, getN), 
  sapply(dadaRs.fr, getN), 
  sapply(dadaFs.rf, getN), 
  sapply(dadaRs.rf, getN), 
  sapply(mergers.fr, getN), 
  sapply(mergers.rf, getN), 
  rowSums(seqtab.nochim), 
  rowSums(seqtab.nochim2)
)
colnames(track) <- c(
  "Demux",
  "Clipped_fr",
  "Clipped_rf", 
  "Filtered_fr", 
  "Filtered_rf", 
  "Denoised_fwd_fr", 
  "Denoised_rev_fr", 
  "Denoised_fwd_rf", 
  "Denoised_rev_rf", 
  "Merged_fr",
  "Merged_rf",
  "Nochim",
  "Tabled"
)
rownames(track) <- c(sample.names)
track <- data.frame(track)


### Write output to file ####
write.table(track, paste0(opt$dir, "/Intermediate_results/", opt$analysis, "/nseqs_all_", opt$name, ".txt"), quote = F, sep = "\t")
write.table(t(seqtab.nochim2), paste0(opt$dir, "/Intermediate_results/", opt$analysis, "/asvtab_", opt$name, ".txt"), quote = F, sep = "\t")
uniquesToFasta(seqtab.nochim2, paste0(opt$dir, "/Intermediate_results/", opt$analysis, "/seqs_", opt$name, ".fasta"))
save.image(paste0(opt$dir, "/Intermediate_results/", opt$analysis, "/workspace_denoising_", opt$name, ".Rdata"))
