#!/usr/bin/env Rscript

# parse assembly metadata for NCBI genomes

### setting up environment ####

# Check if packages are installed
package.list <- c(
  "crayon",
  "optparse",
  "config",
  "dada2",
  "ShortRead",
  "ggplot2",
  "gridExtra",
  "Biostrings",
  "scales",
  "tidyverse",
  "furrr"
)

# Function to check if packages are installed
is.installed <- function(pkg){
  is.element(pkg, installed.packages()[,1])
}

# If not all packages are available
if(any(!is.installed(package.list))) {
  cat("Not all required packages are available. They will now be installed.\n")
  
  # give the user the chance to abort manually
  Sys.sleep(20)
  
  # then install packages
  for(i in which(!is.installed(package.list))) {
    suppressMessages(install.packages(package.list[i], repos = "http://cran.us.r-project.org"))
  }
}

# Break the script if the package installation was unsuccessful
if(any(!is.installed(package.list))) {
  cat(
    paste0(
      "Unable to install all required packages.\nPlease install ",
      paste0(package.list[!is.installed(package.list)], collapse = ", "),
      " manually."
    )
  )
  break
}

# Load packages
cat("Loading libraries...")
silent <- suppressMessages(lapply(package.list, function(X) {require(X, character.only = TRUE)}))
rm(silent)
cat(" done\n")

# Some functions for message output
msg <- function(X){
  cat(crayon::white(paste0("[",format(Sys.time(), "%T"), "]")), X)
}
msg_sub <- function(X){
  cat(crayon::white(paste0("  [",format(Sys.time(), "%T"), "]")), X)
}


### Reading command line options ####

# define command line options
option_list <- list(
  make_option(
    c("-n", "--name"),
    type = "character", 
    default = NULL, 
    help = "project name (will serve as file base name for output files",
    metavar = "character"
  ),
  make_option(
    c("-s", "--samples"),
    type = "character",
    default = NULL,
    help = "file name of sample list (has to be in first column, no headers)",
    metavar = "character"
  ),
  make_option(
    c("-l", "--truncLen_min"),
    type = "integer",
    default = NULL, 
    help = "minimum truncation length for R1 read", 
    metavar = "number"
  ),
  make_option(
    c("-L", "--truncLen_max"),
    type = "integer", 
    default = NULL, 
    help = "maximum truncation length for R1 read",
    metavar = "number"
  ),
  make_option(
    c("-f", "--fragment"),
    type = "integer", 
    default = NULL, 
    help = "maximum size of expected fragment",
    metavar = "number"
  ),
  make_option(
    c("-e", "--error_min"),
    type = "integer", 
    default = NULL, 
    help = "maximum expected error rate (lower bound)",
    metavar = "number"
  ),
  make_option(
    c("-E", "--error_max"),
    type = "integer", 
    default = NULL, 
    help = "maximum expected error rate (upper bound)",
    metavar = "number"
  ),
  make_option(
    c("-d", "--dir"),
    type = "character",
    default = NULL,
    help = "path to analysis directory (highest level directory for sequencing project",
    metavar = "character"
  ),
  make_option(
    c("-a", "--analysis"),
    type = "character",
    default = NULL,
    help = "name of the analysis (version) that is being run",
    metavar = "character"
  ),
  make_option(
    c("-c", "--cpus"),
    type = "integer",
    default = 1, 
    help = "number of threads [default: 1]", 
    metavar = "number"
  )
)

opt_parser <- OptionParser(option_list = option_list)
opt <- parse_args(opt_parser)

if (is.null(opt$name) | is.null(opt$samples) | is.null(opt$truncLen_min) |
    is.null(opt$truncLen_max) | is.null(opt$fragment) | is.null(opt$analysis) |
    is.null(opt$error_min) | is.null(opt$error_max) | is.null(opt$dir)) {
  print_help(opt_parser)
  stop("All parameters are mandatory.\n", call. = FALSE)
}


### Defining functions ####

# Wrapper function to create multi-page pdf with base quality plots
# Argument 'file_base' will be extended by '_R1.pdf' and '_R2.pdf'
quality_check <- function(
  R1, 
  R2, 
  file_base = "QualityProfile",
  cpus = 1
) {

  # start parallel session
  plan(multicore, workers = cpus)

  # R1 profiles
  QualityProfileFs <- future_map(
    1:length(R1),
    function(x) {
      tmp <- list()
      tmp[[1]] <- plotQualityProfile(R1[x])
      return(tmp)
    }
  )
  pdf(paste0(file_base, "_R1.pdf"))
  for(i in 1:length(R1)) {
    do.call("grid.arrange", QualityProfileFs[[i]])  
  }
  dev.off()
  
  # R2 profiles
  QualityProfileRs <- future_map(
    1:length(R2),
    function(x) {
      tmp <- list()
      tmp[[1]] <- plotQualityProfile(R2[x])
      return(tmp)
    }
  )
  pdf(paste0(file_base, "_R2.pdf"))
  for(i in 1:length(R2)) {
    do.call("grid.arrange", QualityProfileRs[[i]])  
  }
  dev.off()
}

# Wrapper function for filtering parameter screening
screen_settings <- function(
  sample.names, 
  fnFs, 
  fnRs, 
  range_maxEE, 
  range_truncLen, 
  cpus = 1
) {
 
  # construct output
  # proportion of total sequences retained, and mean
  # quantiles from min to max in 10% increments
  out <- matrix(NA, ncol = 17, nrow = nrow(range_maxEE) * nrow(range_truncLen))
  colnames(out) <- c(
    "truncLen_R1",
    "truncLen_R2",
    "maxEE_R1",
    "maxEE_R2",
    "prop_total",
    "prop_mean",
    "min",
    paste0("q", seq(10, 90, 10)),
    "max"
  )
  
  # fill first 4 columns
  out[, 1] <- rep(range_truncLen[, 1], each = nrow(range_maxEE))
  out[, 2] <- rep(range_truncLen[, 2], each = nrow(range_maxEE))
  out[, 3] <- rep(range_maxEE[, 1], times = nrow(range_truncLen))
  out[, 4] <- rep(range_maxEE[, 2], times = nrow(range_truncLen))
  
  # create tmp folder for output
  tmp_filtFs <- file.path("tmp", paste0(sample.names, "_F_filt.fastq"))
  tmp_filtRs <- file.path("tmp", paste0(sample.names, "_R_filt.fastq"))
  names(tmp_filtFs) <- sample.names
  names(tmp_filtRs) <- sample.names
  
  # loop through each combination of truncLen and maxEE
  for(i in 1:nrow(out)) {
    msg(paste0("truncLen ", out[i, 1], ",", out[i, 2], " with maxEE ", out[i, 3], ",", out[i, 4], "\n"))
    tmp <- filterAndTrim(
      fwd = fnFs, 
      filt = tmp_filtFs, 
      rev = fnRs, 
      filt.rev = tmp_filtRs,
      truncLen = c(out[i, 1], out[i, 2]),
      maxN = 0,
      minQ = 2,
      maxEE = c(out[i, 3], out[i, 4]), 
      truncQ = 0, 
      rm.phix = TRUE,
      compress = F,
      multithread = cpus
    )
    
    # fill output table
    out[i, "prop_total"] <- round(sum(tmp[, 2])/sum(tmp[, 1]), 4)
    out[i, "prop_mean"] <- round(mean(tmp[, 2]/tmp[, 1]), 4)
    out[i, 7:17] <- round(quantile(tmp[, 2]/tmp[, 1], seq(0, 1, 0.1)), 4)
  }
  
  # remove tmp folder and return output table
  system("rm -rf tmp/")
  return(out)
}


### Defining input and output file paths ####

msg(paste0("This is dada2 version: ", packageVersion("dada2"), ".\n"))

# set sample names
sample.names <- read.table(opt$samples, h = F, sep = "\t", stringsAsFactors = F)[, 1]
msg(paste0("There are ", length(sample.names), " samples in the data set.\n"))

# Specify path to input sequences
path <- paste0(opt$dir, "/Validated_data/", opt$analysis)
fnFs.fr <- paste0(path, "/", sample.names, "_clip_fr_R1.fastq.gz")
fnRs.fr <- paste0(path, "/", sample.names, "_clip_fr_R2.fastq.gz")
fnFs.rf <- paste0(path, "/", sample.names, "_clip_rf_R1.fastq.gz")
fnRs.rf <- paste0(path, "/", sample.names, "_clip_rf_R2.fastq.gz")
names(fnFs.fr) <- sample.names
names(fnRs.fr) <- sample.names
names(fnFs.rf) <- sample.names
names(fnRs.rf) <- sample.names


### Plot quality profiles ####

quality_check(
  c(fnFs.fr, fnFs.rf),
  c(fnRs.fr, fnRs.rf),
  file_base = paste0(opt$dir, "/Intermediate_results/", opt$analysis, "/QualityProfile_", opt$name),
  cpus = opt$cpus
)


### Screen filtering parameters ####

# It is recommended to trim to just enough for the required length for sufficient overlap
# Caution: don't remove too much! I am allowing here for 30bp overlap, although later we will use only 10 in the merging

# Define ranges for truncLen
range_truncLen <- matrix(
  c(
    seq(opt$truncLen_max, opt$truncLen_min, -10),
    opt$fragment - seq(opt$truncLen_max, opt$truncLen_min, -10) + 30
  ),
  nrow = length(seq(opt$truncLen_max, opt$truncLen_min, -10)),
  ncol = 2,
  byrow = F
)

# Define ranges for maxEE
range_maxEE <- matrix(
  c(
    rep(seq(opt$error_min, opt$error_max, 1), each = 2)[-length(rep(seq(opt$error_min, opt$error_max, 1), each = 2))],
    rep(seq(opt$error_min, opt$error_max, 1), each = 2)[-1]
  ),
  nrow = length(seq(opt$error_min, opt$error_max, 1)) * 2 - 1,
  ncol = 2,
  byrow = F
)

msg(paste0("There are ", nrow(range_truncLen) * nrow(range_maxEE), " parameter combinations to be tested.\n"))

# If there are more than 50 sample in the project, randomly pick 50 samples to get a good impression of trimming behavior 
# without having very long run times
if(length(sample.names) > 50) {
  sample.subset <- sample(sample.names, 50)
  msg(paste0("The parameter screening will be run on a subset of samples.\n"))
} else {
  sample.subset <- sample.names
}

# run screening
screen_filt_settings_fr <- screen_settings(
  sample.subset,
  fnFs.fr[sample.subset],
  fnRs.fr[sample.subset],
  range_maxEE,
  range_truncLen,
  cpus = opt$cpus
)
screen_filt_settings_rf <- screen_settings(
  sample.subset,
  fnFs.rf[sample.subset],
  fnRs.rf[sample.subset],
  range_maxEE,
  range_truncLen,
  cpus = opt$cpus
)

# plot screening output
# This is just a gut feeling, but I would optimize for the following criteria:
#   small difference between 10 and 90 percentile of retained reads
#   high total proportion of retained reads
#   most stringent maxEE that does not result in severe loss of reads
pdf(paste0(opt$dir, "/Intermediate_results/", opt$analysis, "/Parameter_screening_", opt$name, ".pdf"), width = 7, height = 7)
plot(
  screen_filt_settings_fr[, "prop_total"],
  screen_filt_settings_fr[, "q90"] - screen_filt_settings_fr[, "q10"],
  col = rep(rainbow(nrow(range_maxEE)), nrow(range_truncLen) - 1),
  pch = 16
)
text(
  screen_filt_settings_fr[, "prop_total"],
  screen_filt_settings_fr[, "q90"] - screen_filt_settings_fr[, "q10"],
  pos = 2,
  col = adjustcolor("black", alpha.f = 0.5)
)
plot(
  screen_filt_settings_rf[, "prop_total"],
  screen_filt_settings_rf[, "q90"] - screen_filt_settings_rf[, "q10"],
  col = rep(rainbow(nrow(range_maxEE)), nrow(range_truncLen) - 1),
  pch = 16
)
text(
  screen_filt_settings_rf[, "prop_total"],
  screen_filt_settings_rf[, "q90"] - screen_filt_settings_rf[, "q10"],
  pos = 2,
  col = adjustcolor("black", alpha.f = 0.5)
)
dev.off()

# write screening output to file
msg("Writing output.\n")
write.table(
  data.frame(
    index = c(1:nrow(screen_filt_settings_fr), 1:nrow(screen_filt_settings_rf)),
    orientation = c(rep("fr", nrow(screen_filt_settings_fr)), rep("rf", nrow(screen_filt_settings_rf))),
    rbind(screen_filt_settings_fr, screen_filt_settings_rf)
  ),
  paste0(opt$dir, "/Intermediate_results/", opt$analysis, "/Parameter_screening_", opt$name, ".txt"),
  sep = "\t",
  quote = F,
  row.names = F
)


