#!/usr/bin/env Rscript

# parse assembly metadata for NCBI genomes

### setting up environment ####

# Check if packages are installed
package.list <- c(
  "crayon",
  "optparse",
  "config"
)

# Function to check if packages are installed
is.installed <- function(pkg){
  is.element(pkg, installed.packages()[,1])
}

# If not all packages are available
if(any(!is.installed(package.list))) {
  cat("Not all required packages are available. They will now be installed.\n")
  break
}

# Load packages
silent <- suppressMessages(lapply(package.list, function(X) {require(X, character.only = TRUE)}))
rm(silent)

# Some functions for message output
msg <- function(X){
  cat(crayon::white(paste0("[",format(Sys.time(), "%T"), "]")), X)
}
msg_sub <- function(X){
  cat(crayon::white(paste0("  [",format(Sys.time(), "%T"), "]")), X)
}


### Reading command line options ####

# define command line options
option_list <- list(
  make_option(
    c("-i", "--input"),
    type = "character",
    default = NULL,
    help = "input file for parsing (subset of blast output for 1 qsedid)",
    metavar = "character"
  ),
  make_option(
    c("-o", "--output"),
    type = "character",
    default = NULL,
    help = "parsed blast output (one line per qseqid)",
    metavar = "character"
  ),
  make_option(
    c("-m", "--acc2tax"),
    type = "character",
    default = NULL,
    help = "accession to taxonomy mapping file (required for parsing blast output",
    metavar = "character"
  ),
  make_option(
    c("-r", "--rank_map"),
    type = "character",
    default = NULL,
    help = "taxon level to rank mapping file for SILVA (required for parsing blast output)",
    metavar = "character"
  )
)

opt_parser <- OptionParser(option_list = option_list)
opt <- parse_args(opt_parser)

if (is.null(opt$input) | is.null(opt$output) |
    is.null(opt$acc2tax) | is.null(opt$rank_map)) {
  print_help(opt_parser)
  stop("All parameters are mandatory.\n", call. = FALSE)
}


### Defining functions ####

# read in functions from for taxpath parsing
SILVAtaxopath <- function(tax, SILVA){
  output <- matrix(NA, nrow = length(tax), ncol = length(levels(SILVA$rank)))
  colnames(output) <- levels(SILVA$rank)
  for (i in 1:length(tax)) {
    for (j in 1:length(levels(SILVA$rank))) {
      if (paste(tax[[i]][1:j], collapse = ";") %in% SILVA$path) {
        output[i, as.character(SILVA[SILVA$path == paste(tax[[i]][1:j], collapse = ";"), "rank"])] <- as.character(SILVA[SILVA$path == paste(tax[[i]][1:j], collapse = ";"), "node"])
      }
    }
  }
  return(output)
}


### Read mapping files ####

silva_acc2tax <- read.table(
  opt$acc2tax,
  h = F,
  sep = "\t",
  stringsAsFactors = F,
  comment.char = "",
  quote = "",
  row.names = 1
)
colnames(silva_acc2tax) <- "tax"
silva <- read.table(
  opt$rank_map,
  h = F,
  sep = "\t",
  stringsAsFactors = F,
  comment.char = "",
  quote = ""
)
silva_map <- data.frame(
  path = gsub(";$", "", silva$V1),
  node = sapply(strsplit(silva$V1, ";"), function(x) x[length(x)]),
  rank = silva$V3,
  stringsAsFactors = T
)


### parse blast output ####

blastout <- read.table(
  opt$input,
  h = F,
  sep = "\t",
  stringsAsFactors = F,
  comment.char = "",
  quote = ""
)
colnames(blastout) <- c(
  "qseqid",
  "sseqid",
  "pident",
  "length",
  "mismatch",
  "gapopen",
  "qstart",
  "qend",
  "sstart",
  "send",
  "evalue",
  "bitscore",
  "qcovs",
  "qlen",
  "gaps"
)

# select best blast hits
tmp.best <- blastout[blastout$bitscore == max(blastout$bitscore), ]
tmp.best <- tmp.best[tmp.best$evalue == min(tmp.best$evalue), ]
tmp.best$taxpath <- silva_acc2tax[tmp.best$sseqid, "tax"]
blastout.best <- tmp.best
rm(tmp.best)

# assign correct taxonomic ranks to each level of the taxonomic path
blastout.best.split <- strsplit(blastout.best$taxpath, ";")
blastout.best.tax <- SILVAtaxopath(blastout.best.split, silva_map)
blastout.best.tax.select <- blastout.best.tax[, c("domain", "phylum", "class", "order", "family", "genus")]

# select most suitable taxonomic path
blastout.best.consensus <- vector("list", length = length(blastout.best))

if(nrow(blastout.best) == 1) {
  tmp <- data.frame(
    blastout.best,
    matrix(
      blastout.best.tax.select,
      nrow = 1,
      ncol = length(blastout.best.tax.select),
      dimnames = list(NULL, names(blastout.best.tax.select))
    ),
    tax_consensus = "unique",
    stringsAsFactors = F
  )
} else {
  tmp.path <- apply(blastout.best.tax.select, 1, paste, collapse = ";")
  if(length(unique(tmp.path)) == 1) {
    tmp <- data.frame(
      blastout.best[1, , drop = F],
      blastout.best.tax.select[1, , drop = F],
      tax_consensus = "unambiguous",
      stringsAsFactors = F
    )
  } else {
    if(sum(table(tmp.path) == max(table(tmp.path))) == 1) {
      tmp <- data.frame(
        blastout.best[which(tmp.path == names(which.max(table(tmp.path))))[1], , drop = F],
        blastout.best.tax.select[which(tmp.path == names(which.max(table(tmp.path))))[1], , drop = F],
        tax_consensus = "frequent",
        stringsAsFactors = F
      )
    } else {
      tmp.tax <- blastout.best.tax.select[tmp.path %in% names(which(table(tmp.path) == max(table(tmp.path)))), ]
      tmp.consensus <- data.frame(
        matrix(
          apply(tmp.tax, 2, function(x) ifelse(length(unique(x)) == 1, unique(x), NA)),
          ncol = ncol(tmp.tax),
          nrow = 1
        ),
        stringsAsFactors = F
      )
      colnames(tmp.consensus) <- colnames(tmp.tax)
      tmp <- data.frame(
        blastout.best[1, , drop = F],
        tmp.consensus,
        tax_consensus = "consensus",
        stringsAsFactors = F
      )
    }
  }
}
blastout.best.consensus <- tmp

if(nrow(blastout.best.consensus) != 1) {
  stop("Something went wrong selecting the consensus taxonomy...\n")
}

# write output file
write.table(
  blastout.best.consensus[, c(1, 14, 3, 13, 17:23), drop = F],
  opt$output,
  quote = F,
  sep = "\t",
  row.names = F,
  col.names = F
)

