#!/usr/bin/env Rscript

# parse assembly metadata for NCBI genomes

### setting up environment ####

# Check if packages are installed
package.list <- c(
  "crayon",
  "optparse",
  "config"
)

# Function to check if packages are installed
is.installed <- function(pkg){
  is.element(pkg, installed.packages()[,1])
}

# If not all packages are available
if(any(!is.installed(package.list))) {
  cat("Not all required packages are available. They will now be installed.\n")
  
  # give the user the chance to abort manually
  Sys.sleep(20)
  
  # then install packages
  for(i in which(!is.installed(package.list))) {
    suppressMessages(install.packages(package.list[i], repos = "http://cran.us.r-project.org"))
  }
}

# Break the script if the package installation was unsuccessful
if(any(!is.installed(package.list))) {
  cat(
    paste0(
      "Unable to install all required packages.\nPlease install ",
      paste0(package.list[!is.installed(package.list)], collapse = ", "),
      " manually."
    )
  )
  break
}

# Load packages
cat("Loading libraries...")
silent <- suppressMessages(lapply(package.list, function(X) {require(X, character.only = TRUE)}))
rm(silent)
cat(" done\n")

# Some functions for message output
msg <- function(X){
  cat(crayon::white(paste0("[",format(Sys.time(), "%T"), "]")), X)
}
msg_sub <- function(X){
  cat(crayon::white(paste0("  [",format(Sys.time(), "%T"), "]")), X)
}


### Reading command line options ####

# define command line options
option_list <- list(
  make_option(
    c("-n", "--name"),
    type = "character", 
    default = NULL, 
    help = "project name (will serve as file base name for output files",
    metavar = "character"
  ),
  make_option(
    c("-l", "--lineage"),
    type = "character",
    default = "Bacteria,Archaea",
    help = "which domains of life were targeted (comma separated)",
    metavar = "character"
  ),
  make_option(
    c("-o", "--blastout"),
    type = "character",
    default = NULL,
    help = "parsed blast output file",
    metavar = "character"
  ),
  make_option(
    c("-f", "--fasta"),
    type = "character",
    default = NULL,
    help = "fasta file with ASVs",
    metavar = "character"
  ),
  make_option(
    c("-d", "--dir"),
    type = "character",
    default = NULL,
    help = "path to analysis directory (highest level directory for sequencing project",
    metavar = "character"
  ),
  make_option(
    c("-v", "--analysis"),
    type = "character",
    default = NULL,
    help = "name of the analysis (version) that is being run",
    metavar = "character"
  )
)

opt_parser <- OptionParser(option_list = option_list)
opt <- parse_args(opt_parser)

if (is.null(opt$name) | is.null(opt$fasta) | is.null(opt$lineage) |
    is.null(opt$blastout) | is.null(opt$dir) | is.null(opt$analysis)) {
  print_help(opt_parser)
  stop("All parameters are mandatory.\n", call. = FALSE)
}


### Defining functions ####

# Further curation of taxonomic paths: removing NAs and uncultured
# append unclassified to last classified level instead of NA
curate_taxpath <- function(tax) {
  out <- tax
  out[out == "uncultured"] <- NA
  k <- ncol(out) - 1
  for (i in 2:k) {
    if (sum(is.na(out[, i])) > 1) {
      test <- out[is.na(out[, i]), ]
      for (j in 1:nrow(test)) {
        if (sum(is.na(test[j, i:(k + 1)])) == length(test[j, i:(k + 1)])) {
          test[j, i] <- paste(test[j, (i - 1)], "_unclassified", sep = "")
          test[j, (i + 1):(k + 1)] <- test[j, i]
        }
      }
      out[is.na(out[, i]), ] <- test
    }
    if (sum(is.na(out[, i])) == 1) {
      test <- out[is.na(out[, i]), ]
      if (sum(is.na(test[i:(k + 1)])) == length(test[i:(k + 1)])) {
        test[i] <- paste(test[(i - 1)], "_unclassified", sep = "")
        test[(i + 1):(k + 1)] <- test[i]
      }
      out[is.na(out[, i]),] <- test
    }
  }
  out[is.na(out[, (k + 1)]), (k + 1)] <- paste(out[is.na(out[, (k + 1)]), k], "_unclassified", sep = "")
  return(out)
}


### curate taxonomic path in parsed blast output ####

blastout <- read.table(
  opt$blastout,
  h = F,
  sep = "\t",
  stringsAsFactors = F,
  comment.char = "",
  quote = ""
)
colnames(blastout) <- c("qseqid", "qlen", "pident", "qcovs", "domain", "phylum", "class", "order", "family", "genus", "tax_consensus")
msg(paste0(nrow(blastout), " ASVs were classified with blast.\n"))

# remove unwanted lineages (see above)
keep_lineage <- strsplit(opt$lineage, ",")[[1]]
blastout.good <- blastout[
  blastout$domain %in% keep_lineage & 
    !grepl("[Cc]hloroplast", blastout$order) &
    !grepl("[Mm]itochondria", blastout$family)
  , ]
msg(paste0("Removing unwanted lineages resulted in ", round(nrow(blastout.good)/nrow(blastout) * 100), "% of ASVs being retained.\n"))

# Further curation of taxonomic paths
# append unclassified to last classified level instead of NA
blastout.clean <- curate_taxpath(blastout.good[, 5:10])

# include sequence information
fasta <- scan(file = opt$fasta, what = "character", sep = "\n")
fasta_df <- data.frame(
  id = gsub("^>", "", fasta[seq(1, length(fasta), 2)]),
  seq = fasta[seq(2, length(fasta), 2)],
  row.names = 1
)

# combine information
blastout.final <- data.frame(
  blastout.good[, c(1:4, 11)],
  blastout.clean,
  stringsAsFactors = F
)
rownames(blastout.final) <- fasta_df[as.character(blastout.final$qseqid), "seq"]

# write output file
write.table(
  blastout.final,
  paste0(opt$dir, "/Intermediate_results/", opt$analysis, "/tax_blast_", opt$name, ".txt"),
  quote = F,
  sep = "\t"
)

